import zipfile
from . import document, meta, tree, user
from .context import context
from sketchformat.layer_group import Page
from sketchformat.serialize import serialize
from typing import Dict, Sequence, List, Tuple, Optional


def convert_fig_tree_to_sketch(
    fig: dict, id_map: Dict[Sequence[int], dict], output: zipfile.ZipFile
) -> None:
    fig_pages, components_page = separate_pages(fig["document"]["children"])

    # We should either bring the fonts to the same indexed_components to pass
    # them as parameter or move the indexed components to the component file
    # and store there the components, for consistency purposes
    if "documentColorProfile" in fig["document"]:
        color_space = fig["document"]["documentColorProfile"]
    else:
        color_space = "sRGB"

    context.init(components_page, id_map, color_space)

    # Convert all normal pages
    sketch_pages: List[Page] = convert_pages(fig_pages, output)

    sketch_document = document.convert(sketch_pages, output)
    sketch_user = user.convert(sketch_pages)
    sketch_meta = meta.convert(sketch_pages)

    write_sketch_file(sketch_document, sketch_user, sketch_meta, output)


def separate_pages(fig_pages: List[dict]) -> Tuple[List[dict], Optional[dict]]:
    components_page = None
    pages = []

    for fig_page in fig_pages:
        if "internalOnly" in fig_page and fig_page["internalOnly"]:
            components_page = fig_page
        else:
            pages.append(fig_page)

    return pages, components_page


def convert_pages(fig_pages: List[dict], output: zipfile.ZipFile) -> List[Page]:
    pages = []

    for fig_page in fig_pages:
        page = tree.convert_node(fig_page, "DOCUMENT")
        serialize(page, output.open(f"pages/{page.do_objectID}.json", "w"))
        pages.append(page)

    if context.symbols_page:
        page = context.symbols_page
        serialize(page, output.open(f"pages/{page.do_objectID}.json", "w"))
        pages.append(page)

    return pages  # type: ignore


def write_sketch_file(
    sketch_document: dict, sketch_user: dict, sketch_meta: dict, output: zipfile.ZipFile
) -> None:
    serialize(sketch_document, output.open("document.json", "w"))
    serialize(sketch_user, output.open("user.json", "w"))
    serialize(sketch_meta, output.open("meta.json", "w"))



import copy
import dataclasses
import math
from .positioning import Vector, Matrix
from converter import utils
from sketchformat.style import *
from typing import List, TypedDict
from .errors import Fig2SketchNodeChanged

BORDER_POSITION = {
    "CENTER": BorderPosition.CENTER,
    "INSIDE": BorderPosition.INSIDE,
    "OUTSIDE": BorderPosition.OUTSIDE,
}

LINE_CAP_STYLE = {
    "NONE": LineCapStyle.BUTT,
    "ROUND": LineCapStyle.ROUND,
    "SQUARE": LineCapStyle.SQUARE,
    "LINE_ARROW": LineCapStyle.SQUARE,
    "ARROW_LINES": LineCapStyle.SQUARE,
    "TRIANGLE_ARROW": LineCapStyle.SQUARE,
    "TRIANGLE_FILLED": LineCapStyle.SQUARE,
}

LINE_JOIN_STYLE = {
    "MITER": LineJoinStyle.MITER,
    "ROUND": LineJoinStyle.ROUND,
    "BEVEL": LineJoinStyle.BEVEL,
}

PATTERN_FILL_TYPE = {
    "STRETCH": PatternFillType.STRETCH,
    "FIT": PatternFillType.FIT,
    "FILL": PatternFillType.FILL,
    "TILE": PatternFillType.TILE,
}

BLEND_MODE = {
    "PASS_THROUGH": BlendMode.NORMAL,
    "NORMAL": BlendMode.NORMAL,
    "DARKEN": BlendMode.DARKEN,
    "MULTIPLY": BlendMode.MULTIPLY,
    # 'LINEAR_BURN': , Unused?
    "COLOR_BURN": BlendMode.COLOR_BURN,
    "LIGHTEN": BlendMode.LIGHTEN,
    "SCREEN": BlendMode.SCREEN,
    # 'LINEAR_DODGE': , Unused?
    "COLOR_DODGE": BlendMode.COLOR_DODGE,
    "OVERLAY": BlendMode.OVERLAY,
    "SOFT_LIGHT": BlendMode.SOFT_LIGHT,
    "HARD_LIGHT": BlendMode.HARD_LIGHT,
    "DIFFERENCE": BlendMode.DIFFERENCE,
    "EXCLUSION": BlendMode.EXCLUSION,
    "HUE": BlendMode.HUE,
    "SATURATION": BlendMode.SATURATION,
    "COLOR": BlendMode.COLOR,
    "LUMINOSITY": BlendMode.LUMINOSITY,
}


def convert(fig_node: dict) -> Style:
    sketch_style = Style(
        do_objectID=utils.gen_object_id(fig_node["guid"], b"style"),
        borderOptions=BorderOptions(
            lineCapStyle=(
                LINE_CAP_STYLE[fig_node["strokeCap"]]
                if "strokeCap" in fig_node
                else BorderOptions.__dict__["lineCapStyle"]
            ),
            lineJoinStyle=(
                LINE_JOIN_STYLE[fig_node["strokeJoin"]]
                if "strokeJoin" in fig_node
                else BorderOptions.__dict__["lineCapStyle"]
            ),
            dashPattern=fig_node.get("dashPattern", []),
        ),
        borders=[convert_border(fig_node, b) for b in fig_node.get("strokePaints", [])],
        fills=[convert_fill(fig_node, f) for f in fig_node.get("fillPaints", [])],
        **convert_effects(fig_node),
        contextSettings=context_settings(fig_node),
    )
    return sketch_style


def convert_border(fig_node: dict, fig_border: dict) -> Border:
    return Border.from_fill(
        convert_fill(fig_node, fig_border),
        position=BORDER_POSITION[fig_node.get("strokeAlign", "CENTER")],
        thickness=fig_node["strokeWeight"],
    )


def convert_fill(fig_node: dict, fig_fill: dict) -> Fill:
    if fig_fill.get("blendMode", "NORMAL") != "NORMAL":
        utils.log_conversion_warning("STY003", fig_node)

    match fig_fill:
        case {"type": "EMOJI"}:
            raise Exception("Unsupported fill: EMOJI")
        case {"type": "SOLID"}:
            # Solid color backgrounds do not support specifying the opacity
            # Instead, it must be set in the color itself
            return Fill.Color(
                convert_color(fig_fill["color"], fig_fill["opacity"]),
                isEnabled=fig_fill["visible"],
            )
        case {"type": "IMAGE"}:
            if is_cropped_image(fig_fill) and not fig_node.get("f2s_cropped_image"):
                # Extract images to a separate layer and retrigger conversion
                convert_crop_image_to_mask(fig_node)

            if "paintFilter" in fig_fill:
                utils.log_conversion_warning("STY005", fig_node)

            return Fill.Image(
                f'images/{fig_fill["image"]["filename"]}',
                patternFillType=PATTERN_FILL_TYPE[fig_fill["imageScaleMode"]],
                patternTileScale=fig_fill.get("scale", 1),
                isEnabled=fig_fill["visible"],
                opacity=fig_fill.get("opacity", 1),
            )
        case _:
            return Fill.Gradient(
                convert_gradient(fig_node, fig_fill),
                isEnabled=fig_fill["visible"],
                opacity=fig_fill.get("opacity", 1),
            )


def is_cropped_image(fig_fill: dict) -> bool:
    return (
        fig_fill.get("type") == "IMAGE"
        and "transform" in fig_fill
        and fig_fill["transform"] != Matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    )


def convert_crop_image_to_mask(fig_node: dict) -> None:
    content = copy.deepcopy(fig_node)
    cropped_images = []
    other_fills = []
    for idx, fill in enumerate(content.get("fillPaints", [])):
        if is_cropped_image(fill):
            image_layer = cropped_image_layer(fig_node, fill)
            image_layer["guid"] = fig_node["guid"] + (1, idx)
            cropped_images.append(image_layer)
        else:
            other_fills.append(fill)

    content["fillPaints"] = other_fills
    content["transform"] = Matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    old = copy.copy(fig_node)
    fig_node.clear()
    fig_node.update(
        {
            "type": "FRAME",
            "name": f'{old["name"]} (crop group)',
            "guid": old["guid"] + (0, 0),
            "size": old["size"],
            "transform": old["transform"],
            "locked": old["locked"],
            "visible": old["visible"],
            "horizontalConstraint": old["horizontalConstraint"],
            "verticalConstraint": old["verticalConstraint"],
            "blendMode": old["blendMode"],
            "opacity": old["opacity"],
            "resizeToFit": True,
            "children": [
                {
                    **content,
                    "mask": True,
                    "maskType": "OUTLINE",
                    "f2s_cropped_image": True,
                },
                *cropped_images,
            ],
        }
    )
    raise Fig2SketchNodeChanged


def cropped_image_layer(fig_node: dict, fill: dict) -> dict:
    invmat = fill["transform"].inv()

    iw = fill["originalImageWidth"]
    ih = fill["originalImageHeight"]
    image_scale = Matrix.scale(1.0 / iw, 1.0 / ih)
    layer_scale = Matrix.scale(fig_node["size"]["x"], fig_node["size"]["y"])

    transform = layer_scale * invmat * image_scale

    height = transform.dot2([0, ih]).length()
    width = transform.dot2([iw, 0]).length()

    normalize_scale = Matrix.scale(iw / width, ih / height)

    image_layer = {
        "type": "RECTANGLE",
        "name": f'{fig_node["name"]} (cropped image)',
        "size": {"x": width, "y": height},
        "transform": transform * normalize_scale,
        "locked": fig_node["locked"],
        "visible": fig_node["visible"],
        "horizontalConstraint": fig_node["horizontalConstraint"],
        "verticalConstraint": fig_node["verticalConstraint"],
        "blendMode": fig_node["blendMode"],
        "opacity": fig_node["opacity"],
        "fillPaints": [fill],
    }
    del image_layer["fillPaints"][0]["transform"]

    return image_layer


def convert_color(color: dict, opacity: Optional[float] = None) -> Color:
    return Color(
        red=color["r"],
        green=color["g"],
        blue=color["b"],
        alpha=color["a"] if opacity is None else opacity,
    )


def convert_gradient(fig_node: dict, fig_fill: dict) -> Gradient:
    # Convert positions depending on the gradient type
    mat = fig_fill["transform"]

    invmat = mat.inv()

    rotation_offset = 0.0
    if fig_fill["type"] == "GRADIENT_LINEAR":
        # Linear gradients always go from (0, .5) to (1, .5)
        # We just apply the transform to get the coordinates (in a 1x1 square)
        return Gradient.Linear(
            from_=Point.from_array(invmat.dot([0, 0.5, 1])),
            to=Point.from_array(invmat.dot([1, 0.5, 1])),
            stops=convert_stops(fig_fill["stops"]),
        )
    elif fig_fill["type"] in ["GRADIENT_RADIAL", "GRADIENT_DIAMOND"]:
        if fig_fill["type"] == "GRADIENT_DIAMOND":
            utils.log_conversion_warning("STY002", fig_node)

        # Angular gradients have the center at (.5, .5), the vertex at (1, .5)
        # and the co-vertex at (.5, 1). We transform them to the coordinates in a 1x1 square
        point_from = invmat.dot([0.5, 0.5, 1])  # Center
        point_to = invmat.dot([1, 0.5, 1])
        point_ellipse = invmat.dot([0.5, 1, 1])

        # Sketch defines the ratio between axis in the item reference point (not the 1x1 square)
        # So we scale the 1x1 square coordinates to fit the ratio of the item frame before
        # calculating the ellipse's ratio
        stroke = fig_node.get("strokeWeight", 0)
        try:
            x_scale = (fig_node["size"]["x"] + 2 * stroke) / (fig_node["size"]["y"] + 2 * stroke)
        except:
            x_scale = 1

        ellipse_ratio = scaled_distance(point_from, point_ellipse, x_scale) / scaled_distance(
            point_from, point_to, x_scale
        )

        return Gradient.Radial(
            from_=Point.from_array(point_from),
            to=Point.from_array(point_to),
            elipseLength=ellipse_ratio,
            stops=convert_stops(fig_fill["stops"]),
        )
    else:
        # Angular gradients don't allow positioning, but we can at least rotate them
        rotation_offset = (
            math.atan2(-fig_fill["transform"][1][0], fig_fill["transform"][0][0]) / 2 / math.pi
        )

        return Gradient.Angular(stops=convert_stops(fig_fill["stops"], rotation_offset))


def convert_stops(fig_stops: List[dict], rotation_offset: float = 0.0) -> List[GradientStop]:
    stops = [
        GradientStop(
            color=convert_color(stop["color"]),
            position=rotated_stop(stop["position"], rotation_offset),
        )
        for stop in fig_stops
    ]

    if rotation_offset:
        # When we have a rotated angular gradient, stops at 0 and 1 both convert
        # to the exact same position and that confuses Sketch. Force a small difference
        stops[-1].position -= 0.00001
    else:
        # Always add a stop at 0 and 1 if needed
        if stops[0].position != 0:
            stops.insert(0, dataclasses.replace(stops[0], position=0))

        if stops[-1].position != 1:
            stops.append(dataclasses.replace(stops[-1], position=1))

    return stops


def scaled_distance(a: Vector, b: Vector, x_scale: float) -> float:
    v = a - b
    return ((v[0] * x_scale) ** 2 + v[1] ** 2) ** 0.5


def rotated_stop(position: float, offset: float) -> float:
    pos = position + offset
    if pos > 1:
        pos -= 1
    elif pos < 0:
        pos += 1
    if pos < 0:
        pos += 1
    return pos


class _Effects(TypedDict):
    blur: Blur
    shadows: List[Shadow]
    innerShadows: List[InnerShadow]


def convert_effects(fig_node: dict) -> _Effects:
    sketch: _Effects = {"blur": Blur.Disabled(), "shadows": [], "innerShadows": []}

    for e in fig_node.get("effects", []):
        if e["type"] == "INNER_SHADOW":
            sketch["innerShadows"].append(
                InnerShadow(
                    blurRadius=e["radius"],
                    offsetX=e["offset"]["x"],
                    offsetY=e["offset"]["y"],
                    spread=e.get("spread", 0),
                    color=convert_color(e["color"]),
                )
            )

        elif e["type"] == "DROP_SHADOW":
            sketch["shadows"].append(
                Shadow(
                    blurRadius=e["radius"],
                    offsetX=e["offset"]["x"],
                    offsetY=e["offset"]["y"],
                    spread=e.get("spread", 0),
                    color=convert_color(e["color"]),
                )
            )

        elif e["type"] == "FOREGROUND_BLUR":
            if sketch["blur"].isEnabled:
                utils.log_conversion_warning("STY001", fig_node)
                continue

            sketch["blur"] = Blur(
                radius=e["radius"] / 2,  # Looks best dividing by 2, no idea why,
                type=BlurType.GAUSSIAN,
            )

        elif e["type"] == "BACKGROUND_BLUR":
            if sketch["blur"].isEnabled:
                utils.log_conversion_warning("STY001", fig_node)
                continue

            sketch["blur"] = Blur(
                radius=e["radius"] / 2,  # Looks best dividing by 2, no idea why,
                type=BlurType.BACKGROUND,
            )

        else:
            raise Exception(f'Unsupported effect: {e["type"]}')

    return sketch


def context_settings(fig_node: dict) -> ContextSettings:
    blend_mode = BLEND_MODE[fig_node["blendMode"]]
    opacity = fig_node["opacity"]

    if fig_node["blendMode"] == "NORMAL" and opacity == 1:
        # Sketch interprets normal at 100% opacity as pass-through
        opacity = 0.99

    return ContextSettings(blendMode=blend_mode, opacity=opacity)



import io
import zipfile
from . import kiwi
from converter.positioning import Matrix
import logging


def decode(path):
    type_converters = {
        "GUID": lambda x: (x["sessionID"], x["localID"]),
        "Matrix": lambda m: Matrix(
            [[m["m00"], m["m01"], m["m02"]], [m["m10"], m["m11"], m["m12"]], [0, 0, 1]]
        ),
    }

    # Open file and check if it's a zip
    fig_zip = None
    reader = open(path, "rb")
    header = reader.read(2)
    reader.seek(0)
    if header == b"PK":
        fig_zip = zipfile.ZipFile(reader)

    try:
        import fig_kiwi

        logging.debug("Using fast (rust) kiwi reader")
        return fig_kiwi.decode(path, type_converters), fig_zip

    except ImportError:
        logging.debug("Falling back to slow (python) kiwi reader")

    if fig_zip:
        reader = fig_zip.open("canvas.fig")

    return kiwi.decode(reader, type_converters), fig_zip



import codecs
import json
from dataclasses import asdict, is_dataclass


def clean_obj(items):
    return {
        # Remove _ suffix for reserved keywords
        k[:-1] if k.endswith("_") else k: v
        for (k, v) in items
        # Skip optional fields that are not present
        if v is not None
    }


def convert_object(obj):
    if is_dataclass(obj):
        return asdict(obj, dict_factory=clean_obj)
    else:
        return obj.to_json()


def serialize(obj, file):
    writer = codecs.getwriter("utf-8")(file)
    json.dump(
        obj,
        writer,
        default=convert_object,
        ensure_ascii=False,  # Write emoji directly, without surrogate pairs
    )



from dataclasses import dataclass, field
from enum import IntEnum
from typing import Optional, List, Union

from .prototype import FlowConnection
from .style import Style, Color


class ExportLayerOptions(IntEnum):
    ALL = 0
    SELECTED = 1
    IN_GROUP = 2


class ExportNamingScheme(IntEnum):
    SUFFIX = 0
    SECONDARY_PREFIX = 1
    PRIMARY_PREFIX = 2


class VisibleScaleType(IntEnum):
    SCALE = 0
    WIDTH = 1
    HEIGHT = 2


class BooleanOperation(IntEnum):
    NONE = -1
    UNION = 0
    SUBTRACT = 1
    INTERSECT = 2
    DIFFERENCE = 3


class LayerListStatus(IntEnum):
    UNDECIDED = 0
    COLLAPSED = 1
    EXPANDED = 2


class ClippingMaskMode(IntEnum):
    OUTLINE = 0
    ALPHA = 1


class ResizeType(IntEnum):
    STRETCH = 0
    PIN_TO_EDGE = 1
    RESIZE = 2
    FLOAT = 2


@dataclass(kw_only=True)
class ExportFormat:
    _class: str = field(default="exportFormat")
    fileFormat: str
    name: str
    visibleScaleType: VisibleScaleType
    absoluteSize: int = 0
    namingScheme: ExportNamingScheme = ExportNamingScheme.PRIMARY_PREFIX
    scale: float = 1


@dataclass(kw_only=True)
class ExportOptions:
    _class: str = field(default="exportOptions")
    exportFormats: List[ExportFormat] = field(default_factory=list)
    includedLayerIds: List[str] = field(default_factory=list)
    layerOptions: ExportLayerOptions = ExportLayerOptions.ALL
    shouldTrim: bool = False


@dataclass(kw_only=True)
class Rect:
    _class: str = field(default="rect")
    height: float
    width: float
    x: float
    y: float
    constrainProportions: bool = False


@dataclass(kw_only=True)
class AbstractLayer:
    do_objectID: str
    frame: Rect
    name: str
    resizingConstraint: int
    rotation: float
    booleanOperation: BooleanOperation = BooleanOperation.NONE
    exportOptions: ExportOptions = field(default_factory=ExportOptions)
    flow: Optional[FlowConnection] = None
    isFixedToViewport: bool = False
    isFlippedHorizontal: bool = False
    isFlippedVertical: bool = False
    isLocked: bool = False
    isTemplate: bool = False
    isVisible: bool = True
    layerListExpandedType: LayerListStatus = LayerListStatus.UNDECIDED
    nameIsFixed: bool = False
    resizingType: ResizeType = ResizeType.STRETCH
    shouldBreakMaskChain: bool = False


@dataclass(kw_only=True)
class AbstractStyledLayer(AbstractLayer):
    style: Style
    hasClippingMask: bool = False
    clippingMaskMode: ClippingMaskMode = ClippingMaskMode.OUTLINE
    sharedStyleID: Optional[str] = None


@dataclass(kw_only=True)
class Slice(AbstractLayer):
    _class: str = field(default="slice")
    hasBackgroundColor: bool = False
    backgroundColor: Color = field(default_factory=Color.White)



import appdirs
import os
import urllib.request
import urllib.parse
import json
import logging
from converter import utils
from fontTools.ttLib import TTFont
from sketchformat.document import FontReference, JsonFileReference
from typing import IO, Tuple
from zipfile import ZipFile

fonts_cache_dir = appdirs.user_cache_dir("Fig2Sketch", "Sketch") + "/fonts"
os.makedirs(fonts_cache_dir, exist_ok=True)


class FontNotFoundError(Exception):
    pass


def retrieve_webfont_family(family):
    WEB_FONT_LIST_URL = "https://fonts.google.com/download/list?family="
    list_url = WEB_FONT_LIST_URL + urllib.parse.quote(family)
    list_response = urllib.request.urlopen(list_url)
    return json.loads(bytearray(list_response.read())[5:])


def get_webfont(family, subfamily):
    family_list = retrieve_webfont_family(family)

    for fi in family_list["manifest"]["fileRefs"]:
        filename = fi["filename"].replace("static/", "")
        if not filename.endswith((".ttf", ".otf")):
            continue
        if family.lower() in filename.lower() and subfamily.lower() in filename.lower():
            font_file_path = f"{fonts_cache_dir}/{filename}"
            if not os.path.exists(font_file_path):
                urllib.request.urlretrieve(fi["url"], font_file_path)

            font_file = open(font_file_path, "rb")
            font_names = extract_names(font_file)

            return font_file, font_names["postscript"]

    raise FontNotFoundError(f"Could not find font {family} {subfamily}")


def convert(
    name: Tuple[str, str], font_file: IO[bytes], postscript: str, output_zip: ZipFile
) -> FontReference:
    family, subfamily = name

    font_file.seek(0)
    data = font_file.read()

    sha = utils.generate_file_ref(data)
    path = f"fonts/{sha}"
    output_zip.open(path, "w").write(data)

    return FontReference(
        do_objectID=utils.gen_object_id((0, 0), bytes.fromhex(sha)),
        fontData=JsonFileReference(_ref_class="MSFontData", _ref=path),
        fontFamilyName=family,
        fontFileName=f"{family}-{subfamily}.ttf",
        postscriptNames=[postscript],
    )


def extract_names(font_file):
    font = TTFont(font_file)
    return {
        "family": font["name"].getBestFamilyName(),
        "subfamily": font["name"].getBestSubFamilyName(),
        "postscript": font["name"].getFirstDebugName((6,)),
    }


def font_matches(fig, font_names):
    return fig["postscript"] == font_names["postscript"] or (
        fig["family"] == font_names["family"] and fig["style"] == font_names["subfamily"]
    )



import logging

try:
    from .orjson import serialize

    logging.debug("Using fast serialization (orjson)")
except:
    from .json import serialize

    logging.debug("Using slow serialization (built-in json)")



from .style import Color
from dataclasses import dataclass, field
from typing import List


@dataclass(kw_only=True)
class JsonFileReference:
    _class: str = field(default="MSJSONFileReference")
    _ref_class: str
    _ref: str


@dataclass(kw_only=True)
class FontReference:
    _class: str = field(default="fontReference")
    do_objectID: str
    fontData: JsonFileReference
    fontFamilyName: str
    fontFileName: str
    postscriptNames: List[str]
    options: int = 3  # Embedded and used


@dataclass(kw_only=True)
class Swatch:
    _class: str = field(default="swatch")
    do_objectID: str
    name: str
    value: Color



from . import base
from sketchformat.layer_common import BooleanOperation
from sketchformat.layer_group import ShapeGroup

BOOLEAN_OPERATIONS = {
    "UNION": BooleanOperation.UNION,
    "INTERSECT": BooleanOperation.INTERSECT,
    "SUBTRACT": BooleanOperation.SUBTRACT,
    "XOR": BooleanOperation.DIFFERENCE,
}


def convert(fig_bool_ops):
    return ShapeGroup(
        **base.base_styled(fig_bool_ops),
    )


def post_process(fig_bool_ops, sketch_bool_ops):
    op = BOOLEAN_OPERATIONS[fig_bool_ops["booleanOperation"]]
    for l in sketch_bool_ops.layers:
        l.booleanOperation = op

    return sketch_bool_ops



from converter import utils
from . import style
from sketchformat.document import Swatch
from typing import Optional


def convert(fig_style: dict) -> Optional[Swatch]:
    match fig_style:
        # Fill with a single fill -> color variable
        case {"styleType": "FILL", "fillPaints": [{"type": "SOLID"} as paint]}:
            uuid = utils.gen_object_id(fig_style["guid"])
            color = style.convert_color(paint["color"], paint["opacity"])
            color.swatchID = uuid
            return Swatch(do_objectID=uuid, name=fig_style["name"], value=color)
        case _:
            return None
